import pyldap
import os
import xml.etree.ElementTree as ET
import platform
from setuphelpers import *

class GpoEngine():

    def __init__(self,domain_name,username=None):
        self.domain_ad = domain_name
        self.site = pyldap.cldap_get_domain_info(domain_name=self.domain_ad)['client_site']

        self.client = pyldap.PyLdapClient(domain_name=self.domain_ad)
        r = self.client.bind_sasl_kerberos()
        self.username = username
        result = self.client.search_all(self.client.root_dn(),"(sAMAccountName=%s)" % self.username)
        parent_ou = list(result)[0]
        parent_ou_split = parent_ou.split(',')
        self.list_ou = [','.join(parent_ou_split[i:]) for i in range(len(parent_ou_split))]

        self.controller = pyldap.cldap_get_ldap_controller(domain_name=self.domain_ad).split(':')[0]
        self.netbios_name = [u for u in next(iter( self.client.search_all(self.client.root_dn(),"(sAMAccountName=%s$)" % self.controller.split('.')[0],['servicePrincipalName']).values())).get('servicePrincipalName')  if len(u.split('/')) > 2 and not '.' in str(u.split('/')[-1])][0].split('/')[-1]
        user_result = self.client.search_all(self.client.root_dn(),"(sAMAccountName=%s)" % self.username,["primaryGroupID","objectSid"])

        self.dn_user = next(iter(user_result))
        self.primaryGroupID = next(iter(user_result.values()))['primaryGroupID'][0]
        self.objectsid_user = next(iter(user_result.values()))['objectSid'][0]

        if platform.system() == 'Darwin':
            self.sysvol = "/private/tmp/sysvol_" + self.username
            if not isdir(self.sysvol):
                mkdirs(self.sysvol)
            run('mount_smbfs //%s/sysvol %s' % (self.controller,self.sysvol))


    def apply_gpo(self,cn_guid):
        guid =cn_guid.split('=')[1].split(',')[0]
        drive_file = r"\\" + self.controller + "\\sysvol\\%s\\Policies\\%s\\User\\Preferences\\Printers\\Printers.xml" % (self.domain_ad,guid)
        return self.parse_drives_xml_with_targeting(drive_file)

    def FilterDomain(self,attrs):
        if attrs['userContext'] == "1":
            return self.netbios_name == attrs['name']
        else:
            return False

    def FilterComputer(self,attrs):
        if attrs['type'] == "NETBIOS":
            return get_hostname().split('.')[0] == attrs['name'].lower()
        if attrs['type'] == "DNS":
            return get_fqdn() == attrs['name'].lower()
        return False

    def FilterUser(self,attrs):
        sidtest= attrs.get('sid')
        if not sidtest:
            return False
        if sidtest == self.objectsid_user:
            return True
        return bool(self.client.search_all(self.client.root_dn(),"(&(sAMAccountName=%s)(objectsid=%s))" % (self.username,sidtest),['objectsid']))

    def FilterGroup(self,attrs):
        sidtest= attrs.get('sid')
        if not sidtest:
            return False

        if attrs['localGroup'] == '1':
            return False

        if attrs['userContext'] != '1':
            return False

        if sidtest.startswith(self.objectsid_user.rsplit('-',1)[0]):
            if str(sidtest.split('-')[-1]) == str(self.primaryGroupID):
                return True
        if attrs['primaryGroup'] == '1':
            return False

        return bool(self.client.search_all(self.client.root_dn(),"(&(member:1.2.840.113556.1.4.1941:=%s)(objectsid=%s))" % (self.dn_user,sidtest),['objectsid']))

    def evaluate_filter(self,filter_element,last_condition_result):
        #on prend que les filtre pris en charge sinon on dit que le résultat c'est non
        type_filter = filter_element.tag
        if (type_filter != 'FilterCollection') and (not type_filter in dir(self)):
            return False

        operator = filter_element.attrib["bool"]

        # si la condition c'est OR et que les dernier check on a résultat a True alors pas besoin de tester
        if operator == "OR" and last_condition_result:
            return True

        #on appel la fonction associter au filter
        if type_filter == "FilterCollection":
            result_filter = self.evaluate_filters(filter_element.findall("*"))
        else:
            result_filter = getattr(self,type_filter)(filter_element.attrib)

        #on vérifie si la condition attendu doit être inversé
        if filter_element.attrib["not"] == '1':
            result_filter = not result_filter

        #On renvoie les resultat en calculant avec les last result
        if operator == "OR":
            return result_filter or last_condition_result

        if operator == "AND":
            return result_filter and last_condition_result

    def evaluate_filters(self,filters_element):

        last_condition_result = None
        last_operator = "ET"

        for filter_element in filters_element:

            #Pour le premier check le last result est True si condition AND, Sinon False
            if last_condition_result == None:
                if filter_element.attrib["bool"] == "OR":
                    last_condition_result = False
                else:
                    last_condition_result = True
            ###################################################

            #On évalue le filter et on passe le resulat précédent pour renvoyer le bon résultat en fonction de l'operateur
            result_filter = self.evaluate_filter(filter_element,last_condition_result)

            # comportement étrange mais visiblement le moteur de ciblage fonctionne comme ceci actuellement
            # Si le dernier résultat est True et que le dernier Operateur a True et l'operateur actuelle est a True alors on s'arrête
            if last_condition_result and last_operator == "OR" and  filter_element.attrib["bool"] == "OR":
                return True
            ####################################################################

            # on conserve les resultat pour le prochaine check
            last_condition_result = result_filter
            last_operator = filter_element.attrib["bool"]

        return last_condition_result


    def parse_drives_xml_with_targeting(self,xml_file_path):
        """
        Parse the XML file and extract the 'path' properties of all drives,
        including targeting conditions described in the Drive Map documentation.

        Args:
            xml_file_path (str): Path to the XML file.

        Returns:
            dict: A dictionary of share paths indexed by drive letters, including targeting conditions.
        """
        try:
            if platform.system() == 'Windows':

                with open(xml_file_path, 'r', encoding='utf-8') as f:
                    xml_data = f.read()

            else:
                path_parts = xml_file_path.split("\\",4)

                if platform.system() == 'Linux':
                    srv = path_parts[2]
                    share = path_parts[3]
                    subpath = path_parts[4].rsplit('\\',1)[0]
                    file = path_parts[4].rsplit('\\',1)[1]
                    xml_data = run(f'smbclient -N --use-kerberos=desired //{srv}/{share} -D "{subpath}" -c "get {file} /dev/stdout"').rsplit('getting file ',1)[0]

                if platform.system() == 'Darwin':
                    with open(self.sysvol + '/' + '/'.join(path_parts[-1].split('\\')), 'r', encoding='utf-8') as f:
                        xml_data = f.read()

            # Analyser le contenu XML
            root = ET.fromstring(xml_data)

            # Dictionnaire pour stocker les lecteurs et leurs propriétés
            share_paths = {}

            for drive in root.findall("SharedPrinter"):

                if drive.attrib.get('disabled') == "1":
                    continue

                properties = drive.find("Properties")

                if properties is not None:
                    # Extraire les propriétés générales
                    props = {
                        key: properties.get(key)
                        for key in ['path', 'useLetter', 'label', 'action',
                                    'thisDrive', 'allDrives', 'userName', 'letter']
                        if properties.get(key)
                    }
                    if drive.find("Filters"):
                        if not self.evaluate_filters(drive.find("Filters")):
                            continue

                    # Utiliser la lettre comme clé principale
                    if "path" in props:
                        share_paths[props["path"]] = props

            return share_paths

        except FileNotFoundError:
            print(f"Erreur : Le fichier '{xml_file_path}' n'existe pas.")
            return {}
        except ET.ParseError:
            print("Erreur : Impossible d'analyser le fichier XML. Vérifiez sa validité.")
            return {}

    def check_acl_gpo(self,data):
        datasplit = data.split('(')

        group_primary =  self.objectsid_user.rsplit('-',1)[0] + '-' + str(self.primaryGroupID)

        all_sid = {}
        for p in datasplit:
            if ';edacfd8f-ffb3-11d1-b41d-00a0c968f939;' in p:
                sid_test = p.split(';')[5].split(')')[0]
                all_sid[sid_test] = None

        all_user_sid_member = {group_primary:None,"AU":None}

        if all_sid:
            filter_ldap = "(&(|%s)(member:1.2.840.113556.1.4.1941:=%s))" %  (''.join(['(objectsid=' + u + ')' for u in all_sid  ]),self.dn_user)
            result_ldap = self.client.search_all(self.client.root_dn(),filter_ldap,['objectSid'])
            for entry in result_ldap:
                all_user_sid_member[result_ldap[entry]['objectSid'][0]] = None


        for p in datasplit:
            if ';edacfd8f-ffb3-11d1-b41d-00a0c968f939;' in p:
                if not p.split(';')[0] == 'OD':
                    continue
                sid_test = p.split(';')[5].split(')')[0]
                if sid_test in all_user_sid_member:
                    return False

        for p in datasplit:
            if ';edacfd8f-ffb3-11d1-b41d-00a0c968f939;' in p:
                if not p.split(';')[0] == 'OA':
                    continue
                sid_test = p.split(';')[5].split(')')[0]
                if sid_test in all_user_sid_member:
                    return True

        return False

    def get_printers(self):
        #search all OU
        dict_network_drive={}

        dict_ou_result={}
        for ou in self.list_ou:
            dict_ou_result[ou]= self.client.search_all(self.client.root_dn(),"(&(distinguishedName=%s)(gPLink=*)(!(gPOptions=2))(!(gPOptions=3)))" % ou,['gPOptions','gPLink']).get(ou)

        if self.site :
            rsite = self.client.search_all('CN=Configuration,' + self.client.root_dn(), '(&(name=%s)(objectClass=site)(gPLink=*))' % self.site)
            if rsite:
                dict_ou_result[self.list_ou[-1]]=list(rsite.values())[0]

        #on cherche si il y a une ou avec block inheritance
        block_inheritance=None
        for ou in self.list_ou:
            if not dict_ou_result.get(ou):
                continue
            if dict_ou_result[ou].get('gPOptions',[''])[0] == 1:
                block_inheritance = ou.lower()
                break

        dict_gpo_value={}

        #première pass
        for ou in reversed(self.list_ou):

            # si l'OU que je traite ne termine pas par l'ou block inheritance
            if block_inheritance:
                if not ou.lower().endswith(block_inheritance):
                    continue

            if not dict_ou_result.get(ou):
                continue

            for u in dict_ou_result[ou]['gPLink'][0].split('['):
                if not u :
                    continue

                if not u in dict_gpo_value:
                    if not u.strip():
                        continue
                    gpo_value = list(self.client.search_all(self.client.root_dn(),"(distinguishedName=%s)" % u.split('//')[1].rsplit(';')[0],['nTSecurityDescriptor','gPCUserExtensionNames','flags']).values())[0]
                    dict_gpo_value[u] = gpo_value
                else:
                    gpo_value = dict_gpo_value[u]

                ## 0 Aucun drapeau actif (configuration par défaut).
                ## 1 La GPO est désactivée côté utilisateur.
                ## 2 La GPO est désactivée côté ordinateur.
                ## 3 La GPO est désactivée à la fois pour les utilisateurs et ordinateurs.
                ## 4 Indique que la GPO est une stratégie de groupe starter (modèle de GPO préconfiguré).

                if gpo_value['flags'][0] in [1,3]:
                    continue

                # si ce n'est pas une gpo network drive on fait rien
                if not '{A8C42CEA-CDB8-4388-97F4-5831F933DA84}' in str(gpo_value.get('gPCUserExtensionNames','')):
                    continue

                gpo_param= u.split('//')[1].rsplit(';',1)
                if gpo_param[1].strip(']') == "0":
                    if not self.check_acl_gpo(gpo_value['nTSecurityDescriptor'][0]):
                        continue
                    dict_network_drive.update(self.apply_gpo(gpo_param[0]))

        #on repasse dessu pour appliquer les gpo "appliqué"
        print('----------------------')
        for ou in reversed(self.list_ou):
            if not dict_ou_result.get(ou):
                continue
            for u in dict_ou_result[ou]['gPLink'][0].split('['):
                if not u :
                    continue
                if not u in dict_gpo_value:
                    if not u.strip():
                        continue
                    gpo_value = list(self.client.search_all(self.client.root_dn(),"(distinguishedName=%s)" % u.split('//')[1].rsplit(';')[0],['nTSecurityDescriptor','gPCUserExtensionNames','flags']).values())[0]
                    dict_gpo_value[u] = gpo_value
                else:
                    gpo_value = dict_gpo_value[u]

                ## 0 Aucun drapeau actif (configuration par défaut).
                ## 1 La GPO est désactivée côté utilisateur.
                ## 2 La GPO est désactivée côté ordinateur.
                ## 3 La GPO est désactivée à la fois pour les utilisateurs et ordinateurs.
                ## 4 Indique que la GPO est une stratégie de groupe starter (modèle de GPO préconfiguré).

                if gpo_value['flags'][0] in [1,3]:
                    continue

                # si ce n'est pas une gpo network drive on fait rien
                if not '{A8C42CEA-CDB8-4388-97F4-5831F933DA84}' in str(gpo_value.get('gPCUserExtensionNames','')):
                    continue

                gpo_param= u.split('//')[1].rsplit(';',1)
                #0 case appliqué décocher, lien activé cocher
                #1 case appliqué décocher, lien activé décocher
                #2 case appliqué cocher, lien activé cocher
                #3 case appliqué cocher, lien activé décocher
                if gpo_param[1].strip(']') == "2":
                    if not self.check_acl_gpo(gpo_value['nTSecurityDescriptor'][0]):
                        continue
                    dict_network_drive.update(self.apply_gpo(gpo_param[0]))

        return dict_network_drive

def install():
    pass

def session_setup():
    run('/opt/wapt/wapt-get.bin audit %s -f -S' % control.package)

def audit():

    dict_print_user={}
    for uc in get_loggedinusers():
        ge=GpoEngine(domain_name=get_domain_hostname_from_keytab(),username=uc)
        list_printer = ge.get_printers()
        for printer in list_printer:
            if printer in dict_print_user:
                dict_print_user[printer].append(uc)
            else:
                dict_print_user[printer] = [uc]


    for p in dict_print_user:
        name_printer = p.split('\\')[-1]
        urlprinter = 'smb:' + p.replace('\\','/')
        run('/usr/sbin/lpadmin -p "%s" -v "%s" -o auth-info-required=negotiate -E -u allow:%s' % (name_printer,urlprinter,','.join(dict_print_user[p])))